#include "msg.h"
#include "tbx.h"
#include "sched_struct.h"
#define VERBOSE
#include "messages.h"

float task_computation_amount = 1000;
float task_communication_amount = 10;
float request_computation_amount = 0.1;
float request_communication_amount = 0.1;
float global_time = 0;
int heuristic = 0;

typedef enum {
  FASTEST_FIRST=1,
} HEURISTICS;

m_task_t work_task_create(void)
{
  static int number = 0;
  char buffer[128];

  sprintf(buffer, "Work Task (%d)", number++);

  return MSG_task_create(buffer, task_computation_amount,
			 task_communication_amount, NULL);
}

void work_task_destroy(m_task_t * task)
{
  MSG_task_destroy(*task);
  *task = MSG_TASK_UNINITIALIZED;
}

m_task_t request_task_create(m_host_t who_am_i)
{
  static int number = 0;
  char buffer[128];
  request_data_t data = calloc(1, sizeof(struct s_request_data));

  sprintf(buffer, "Request Task (%d)", number++);

  data->who_am_i = who_am_i;

  return MSG_task_create(buffer, request_computation_amount,
			 request_communication_amount, data);
}

void request_task_destroy(m_task_t * task)
{
  free((*task)->data);
  MSG_task_destroy(*task);
  *task = MSG_TASK_UNINITIALIZED;
}

void sched_master(int argc, char *argv[])
{
  int nb_task = 0;
  TBX_FIFO_t task_list = TBX_FIFO_newFIFO();

  {                  /* Creating all the tasks to proces */
    int i;
    ASSERT((argc==4),"Usage: sched_master <number_of_tasks> <task_computation> <task_communication>");

    nb_task = atoi(argv[1]);
    task_computation_amount = atof(argv[2]);
    task_communication_amount = atof(argv[3]);

    for (i = 0; i < nb_task; i++) {
      TBX_FIFO_insert(task_list, work_task_create());
    }
  }

  
  PRINT_MESSAGE("Got %d task to process.\n",
		TBX_FIFO_size(task_list));
  PRINT_MESSAGE("Waiting for some slave to contact me.\n");

  while (TBX_FIFO_size(task_list) > 0) {
    m_task_t task = TBX_FIFO_extract(task_list);
    m_task_t request = MSG_TASK_UNINITIALIZED;
    m_host_t slave;
    float power;
    float bandwidth;

    MSG_task_get(&request, REQUEST_CHANNEL);
    MSG_task_execute(request);

    slave = ((request_data_t) request->data)->who_am_i;
    power = ((request_data_t) request->data)->computing_power;
    bandwidth = ((request_data_t) request->data)->bandwidth;

    PRINT_MESSAGE
	("Received \"%s\" from \"%s\". Here is its computing power : %f and its bandwidth : %f \n",
	 request->name, MSG_host_get_name(slave), power);
    request_task_destroy(&request);

    MSG_task_put(task, slave, GENERAL_CHANNEL);
  }

  PRINT_MESSAGE("No more Task to process. Bye!\n");
  return;
}

void sched_slave(int argc, char *argv[])
{
  m_host_t father = NULL;

  ASSERT((argc==2),"Usage: sched_slave <master_name>");
  ASSERT(((father=MSG_get_host_by_name(argv[1]))!=NULL),"Unknown host");

  while (1) {
    m_task_t request = request_task_create(MSG_host_self());
    m_task_t task = MSG_TASK_UNINITIALIZED;

    PRINT_MESSAGE
	("I have nothing to do. Let's ask some work to my father\n");
    MSG_task_put(request, father, REQUEST_CHANNEL);
    MSG_task_get(&task, GENERAL_CHANNEL);
    PRINT_MESSAGE("Great ! Some Work ! Processing %s\n", task->name);
    MSG_task_execute(task);
    work_task_destroy(&task);
  }
  return;
}

int by_power(const void *a, const void *b) 
{
  float power_a = (((request_data_t) (*((m_task_t*)a))->data))->computing_power;
  float power_b = (((request_data_t) (*((m_task_t*)b))->data))->computing_power;
  if(power_a>=power_b) return -1;
  if(power_a==power_b) return 0;
  return 1;
}

void* extract_fastest(TBX_FIFO_t request_list) 
{
  int i;
  int n=TBX_FIFO_size(request_list);
  void **tab = CALLOC(n,sizeof(void*));

  for(i=0;i<n;i++) {
    tab[i] = TBX_FIFO_extract(request_list);
  }
  qsort(tab,n,sizeof(void*),by_power);
  for(i=0;i<n;i++) {
    TBX_FIFO_insert(request_list,tab[i]);
  }
  
  return (TBX_FIFO_extract(request_list));
}

void sched_wiser_master(int argc, char *argv[])
{
  int nb_task = 0;
  TBX_FIFO_t task_list = TBX_FIFO_newFIFO();
  int max_request_number = 1;
  TBX_FIFO_t pending_request = TBX_FIFO_newFIFO();

  {                  /* Creating all the tasks to proces */
    int i;
    ASSERT((argc==6),"Usage: sched_wiser_master <number_of_tasks> <task_computation> <task_communication> <heuristic_nb> <max_pending_request>");

    nb_task = atoi(argv[1]);
    task_computation_amount = atof(argv[2]);
    task_communication_amount = atof(argv[3]);
    heuristic = atoi(argv[4]);
    max_request_number = atoi(argv[5]);

    for (i = 0; i < nb_task; i++) {
      TBX_FIFO_insert(task_list, work_task_create());
    }
  }
  
  PRINT_MESSAGE("Got %d task to process.\n",
		TBX_FIFO_size(task_list));
  PRINT_MESSAGE("Waiting for some slave to contact me.\n");

  while (TBX_FIFO_size(task_list) > 0) {
    if ((TBX_FIFO_size(pending_request) < max_request_number)) {
      m_task_t request = MSG_TASK_UNINITIALIZED;
      MSG_task_get(&request, REQUEST_CHANNEL);
      MSG_task_execute(request);
      TBX_FIFO_insert(pending_request, request);
    } else {
      m_task_t request = MSG_TASK_UNINITIALIZED;
      m_task_t task = TBX_FIFO_extract(task_list);
      m_host_t slave;
      float power, bandwidth;

      switch(heuristic) {
      case FASTEST_FIRST: { 
	request = extract_fastest(pending_request); break;
      }
      default: {
	request = TBX_FIFO_extract(pending_request);
      }
      }

      slave = ((request_data_t) request->data)->who_am_i;
      power = ((request_data_t) request->data)->computing_power;
      bandwidth = ((request_data_t) request->data)->bandwidth;

      PRINT_MESSAGE
	  ("Received \"%s\" from \"%s\". Here is its computing power : %f and its bandwidth : %f \n",
	   request->name, MSG_host_get_name(slave), power, bandwidth);
      request_task_destroy(&request);

      task->data = calloc(1, sizeof(float));
      (*((float *) task->data)) = ((float) MSG_getClock());
      MSG_task_put(task, slave, GENERAL_CHANNEL);
    }
  }

  PRINT_MESSAGE("No more Task to process. Bye!\n");
  return;
}

void sched_wiser_slave(int argc, char *argv[])
{
  float processing_time = 10.0;
  float transfer_time = 0.02;
  m_host_t father = NULL;

  ASSERT((argc==2),"Usage: sched_wiser_slave <master_name>");
  ASSERT(((father=MSG_get_host_by_name(argv[1]))!=NULL),"Unknown host");

  while (1) {
    m_task_t request = request_task_create(MSG_host_self());
    m_task_t task = MSG_TASK_UNINITIALIZED;

    ((request_data_t) request->data)->computing_power =
	1 / processing_time;
    ((request_data_t) request->data)->bandwidth = 1 / transfer_time;
    PRINT_MESSAGE
	("I have nothing to do. Let's ask some work to my father\n");
    MSG_task_put(request, father, REQUEST_CHANNEL);
    transfer_time = MSG_getClock();
    MSG_task_get(&task, GENERAL_CHANNEL);
    transfer_time =
	MSG_getClock() - MAX(transfer_time, (*((float *) task->data)));
    PRINT_MESSAGE("Great ! Some Work ! Processing %s\n", task->name);
    processing_time = MSG_getClock();
    MSG_task_execute(task);
    free(task->data);
    work_task_destroy(&task);
    processing_time = MSG_getClock() - processing_time;
  }
  return;
}
